package com.yourkit.probes.builtin;

import com.yourkit.probes.*;

/**
 * Detect long running AWT/Swing events
 */
public final class AWTEvents {

  private static final class EventTable extends Table {
    private final StringColumn myRunnableClass = new StringColumn("Event Class");

    public EventTable() {
      super(
        "Long AWT Events",
        // It makes no sense to record stack trace

        Table.LASTING_EVENTS |
        Table.RECORD_THREAD_CPU_TIME |
        Table.RECORD_THREAD_ON_ROW_CREATION
      );
    }
  }
  private static final EventTable T_EVENT = new EventTable();
  static {
    // Record only events longer than 100 ms
    T_EVENT.setMinimumRecordedLastingEventTime(300);

    // If event lasts long enough, start collecting stack traces more frequently,
    // to learn what is happening.
    // Note, that the threshold of setFrequentSamplingThreshold() is slightly smaller
    // than the threshold passed to setMinimumRecordedLastingEventTime(),
    // to ensure that every retained event has associated stack traces, thus enabling its analysis.
    T_EVENT.setFrequentSamplingThreshold(250);
  }

  @MethodPattern("java.awt.EventQueue:dispatchEvent(java.awt.AWTEvent)")
  public static final class EventQueue_dispatchEvent_Probe {
    public static int onEnter() {
      return T_EVENT.createRow();
    }

    public static void onReturn(
      final @OnEnterResult int rowIndex,
      final @Param(1) Object event
    ) {
      processExitImpl(rowIndex, event);
    }

    public static void onUncaughtException(
      final @OnEnterResult int rowIndex,
      final @Param(1) Object event
    ) {
      processExitImpl(rowIndex, event);
    }

    /**
     * Close event whenever it normally finished or terminated by an exception
     */
    private static void processExitImpl(final int rowIndex, final Object event) {
      if (rowIndex < 0) {
        return;
      }

      if (!T_EVENT.closeRow(rowIndex)) {
        // optimization: short event not stored in table - do not set its values
        return;
      }

      T_EVENT.myRunnableClass.setValue(rowIndex, event.getClass().getName());
    }
  }
}
