package com.yourkit.probes.builtin;

import com.yourkit.probes.MethodPattern;
import com.yourkit.probes.OnEnterResult;
import com.yourkit.probes.ReturnValue;
import com.yourkit.probes.This;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.util.WeakHashMap;

public final class FileChannels {

  interface ChannelHandler {
    int readOnEnter();
    void readOnReturn(int rowIndex, long bytesRead);

    int writeOnEnter();
    void writeOnReturn(int rowIndex, long bytesWritten);

    int transferToOnEnter();
    void transferToOnReturn(int rowIndex, long transferedBytes);

    int transferFromOnEnter();
    void transferFromOnReturn(int rowIndex, long transferedBytes);
  }

  private static final WeakHashMap<FileChannel, ChannelHandler> ourFileChannel2ChannelHandler = new WeakHashMap<FileChannel, ChannelHandler>();

  @MethodPattern(
    {
      "java.io.RandomAccessFile:getChannel()",
      "java.io.FileInputStream:getChannel()",
      "java.io.FileOutputStream:getChannel()"
    }
  )
  public static final class GetChannel_Probe {
    public static void onReturn(
      final @ReturnValue FileChannel fileChannel,
      final @This Object _this
    ) {
      final ChannelHandler channelHandler;
      if (_this instanceof FileInputStream) {
        channelHandler = FileStreams.registerInputChannel((FileInputStream)_this);
      }
      else if (_this instanceof FileOutputStream) {
        channelHandler = FileStreams.registerOutputChannel((FileOutputStream)_this);
      }
      else {
        channelHandler = RandomAccessFiles.registerChannel((RandomAccessFile)_this);
      }
      if (channelHandler == null) {
        return;
      }
      synchronized (ourFileChannel2ChannelHandler) {
        ourFileChannel2ChannelHandler.put(fileChannel, channelHandler);
      }
    }
  }

  @MethodPattern(
    "sun.nio.ch.FileChannelImpl:transferFrom(java.nio.channels.ReadableByteChannel, long, long)"
  )
  public static final class FileChannel_transferFrom_Probe {
    public static int onEnter(
      final @This FileChannel fileChannel
    ) {
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return -1;
      }
      return channelHandler.transferFromOnEnter();
    }

    public static void onReturn(
      final @OnEnterResult int transferRowIndex,
      final @ReturnValue long transferedBytes,
      final @This FileChannel fileChannel
    ) {
      if (transferRowIndex == -1) {
        return;
      }
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return;
      }
      channelHandler.transferFromOnReturn(transferRowIndex, transferedBytes);
    }
  }

  @MethodPattern(
    "sun.nio.ch.FileChannelImpl:transferTo(long, long, java.nio.channels.WritableByteChannel)"
  )
  public static final class FileChannel_transferTo_Probe {
    public static int onEnter(
      final @This FileChannel fileChannel
    ) {
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return -1;
      }
      return channelHandler.transferToOnEnter();
    }

    public static void onReturn(
      final @OnEnterResult int transferRowIndex,
      final @ReturnValue long transferedBytes,
      final @This FileChannel fileChannel
    ) {
      if (transferRowIndex == -1) {
        return;
      }
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return;
      }
      channelHandler.transferToOnReturn(transferRowIndex, transferedBytes);
    }
  }

  @MethodPattern(
    {
      "sun.nio.ch.FileChannelImpl:write(java.nio.ByteBuffer)" /* returns int */,
      "sun.nio.ch.FileChannelImpl:write(java.nio.ByteBuffer, long)" /* returns int */
    }
  )
  public static final class FileChannel_write_Probe1 {
    public static int onEnter(
      final @This FileChannel fileChannel
    ) {
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return -1;
      }
      return channelHandler.writeOnEnter();
    }

    public static void onReturn(
      final @OnEnterResult int writeRowIndex,
      final @ReturnValue int bytesWritten,
      final @This FileChannel fileChannel
    ) {
      if (writeRowIndex == -1) {
        return;
      }
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return;
      }
      channelHandler.writeOnReturn(writeRowIndex, bytesWritten);
    }
  }

  @MethodPattern(
    "sun.nio.ch.FileChannelImpl:write(java.nio.ByteBuffer[], int, int)" /* returns long */
  )
  public static final class FileChannel_write_Probe2 {
    public static int onEnter(
      final @This FileChannel fileChannel
    ) {
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return -1;
      }
      return channelHandler.writeOnEnter();
    }

    public static void onReturn(
      final @OnEnterResult int writeRowIndex,
      final @ReturnValue long bytesWritten,
      final @This FileChannel fileChannel
    ) {
      if (writeRowIndex == -1) {
        return;
      }
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return;
      }
      channelHandler.writeOnReturn(writeRowIndex, bytesWritten);
    }
  }

  @MethodPattern(
    {
      "sun.nio.ch.FileChannelImpl:read(java.nio.ByteBuffer)" /* returns int */,
      "sun.nio.ch.FileChannelImpl:read(java.nio.ByteBuffer, long)" /* returns int */
    }
  )
  public static final class FileChannel_read_Probe1 {
    public static int onEnter(
      final @This FileChannel fileChannel
    ) {
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return -1;
      }
      return channelHandler.readOnEnter();
    }

    public static void onReturn(
      final @OnEnterResult int readRowIndex,
      final @ReturnValue int bytesRead,
      final @This FileChannel fileChannel
    ) {
      if (readRowIndex == -1) {
        return;
      }
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return;
      }
      channelHandler.readOnReturn(readRowIndex, bytesRead);
    }
  }

  @MethodPattern(
    "sun.nio.ch.FileChannelImpl:read(java.nio.ByteBuffer[], int, int)" /* returns long */
  )
  public static final class FileChannel_read_Probe2 {
    public static int onEnter(
      final @This FileChannel fileChannel
    ) {
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return -1;
      }
      return channelHandler.readOnEnter();
    }

    public static void onReturn(
      final @OnEnterResult int readRowIndex,
      final @ReturnValue long bytesRead,
      final @This FileChannel fileChannel
    ) {
      if (readRowIndex == -1) {
        return;
      }
      final ChannelHandler channelHandler;
      synchronized (ourFileChannel2ChannelHandler) {
        channelHandler = ourFileChannel2ChannelHandler.get(fileChannel);
      }
      if (channelHandler == null) {
        return;
      }
      channelHandler.readOnReturn(readRowIndex, bytesRead);
    }
  }
}
