package com.yourkit.probes.builtin;

import com.yourkit.probes.*;

import javax.servlet.Filter;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.jsp.HttpJspPage;
import java.lang.reflect.Method;

public final class Servlets {
  private static final class RequestTable extends Table {
    private final StringColumn myURI = new StringColumn("URI");
    private final StringColumn myQuery = new StringColumn("Query");
    private final StringColumn myInclude = new StringColumn("Include");

    public RequestTable() {
      super("JSP/Servlet Requests", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final RequestTable T_REQUEST = new RequestTable();

  @MethodPattern("*:service(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)")
  public static final class Servlet_service_Probe {
    public static int onEnter(
      final @This HttpServlet servlet,
      final @Param(1) HttpServletRequest request
    ) {
      final String servletClassName = servlet.getClass().getName();
      if (
        servletClassName.equals("javax.servlet.http.HttpServlet") ||
        servletClassName.startsWith("org.apache.jasper.")
      ) {
        // ignore calls from standard classes; they will delegate to user servlets
        return -1;
      }

      // ignore calls to Servlet's service() for jsp pages
      try {
        final Class<?> jspPageClass = Class.forName("javax.servlet.jsp.HttpJspPage", false, request.getClass().getClassLoader());
        if (jspPageClass.isInstance(servlet)) {
          return -1;
        }
      }
      catch (final Throwable e) {
      }

      final int rowIndex = T_REQUEST.createRow();
      T_REQUEST.myURI.setValue(rowIndex, getRequestURI(request));
      T_REQUEST.myQuery.setValue(rowIndex, getQuery(request));
      T_REQUEST.myInclude.setValue(rowIndex, getRequestAttribute(request, "javax.servlet.include.request_uri"));
      return rowIndex;
    }

    public static void onReturn(final @OnEnterResult int rowIndex) {
      if (rowIndex == -1) {
        return;
      }

      T_REQUEST.closeRow(rowIndex);
    }
  }

  @MethodPattern("*:_jspService(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)")
  public static final class HttpJspPage_jspService_Probe {
    public static int onEnter(
      final @This HttpJspPage jspPage,
      final @Param(1) HttpServletRequest request
    ) {
      final int rowIndex = T_REQUEST.createRow();
      T_REQUEST.myURI.setValue(rowIndex, getRequestURI(request));
      T_REQUEST.myQuery.setValue(rowIndex, getQuery(request));
      T_REQUEST.myInclude.setValue(rowIndex, getRequestAttribute(request, "javax.servlet.include.request_uri"));
      return rowIndex;
    }

    public static void onReturn(final @OnEnterResult int rowIndex) {
      T_REQUEST.closeRow(rowIndex);
    }
  }

  @MethodPattern("*:doFilter(javax.servlet.ServletRequest, javax.servlet.ServletResponse, javax.servlet.FilterChain)")
  public static final class Filter_doFilter_Probe {
    public static int onEnter(
      final @This Filter filter,
      final @Param(1) ServletRequest request
    ) {
      final int rowIndex = T_REQUEST.createRow();
      T_REQUEST.myURI.setValue(rowIndex, getRequestURI(request));
      T_REQUEST.myQuery.setValue(rowIndex, getQuery(request));
      T_REQUEST.myInclude.setValue(rowIndex, getRequestAttribute(request, "javax.servlet.include.request_uri"));
      return rowIndex;
    }

    public static void onReturn(final @OnEnterResult int rowIndex) {
      if (rowIndex == -1) {
        return;
      }
      T_REQUEST.closeRow(rowIndex);
    }
  }

  private static String getQuery(final Object request) {
    try {
      final Class requestClass = request.getClass();
      final Method method = requestClass.getDeclaredMethod("getQueryString");
      method.setAccessible(true);
      return (String)method.invoke(request);
    }
    catch(final Throwable e) {
      return "<unknown>";
    }
  }

  /**
   * This method is implemented using reflection to avoid class-loading issues
   */
  private static String getRequestURI(final Object request) {
    try {
      final Class requestClass = request.getClass();
      final Method getRequestURIMethod = requestClass.getDeclaredMethod("getRequestURI");
      getRequestURIMethod.setAccessible(true);
      return (String)getRequestURIMethod.invoke(request);
    }
    catch(final Throwable e) {
      return "<unknown>";
    }
  }

  private static String getRequestAttribute(final Object request, final String attributeName) {
    try {
      final Class requestClass = request.getClass();
      final Method method = requestClass.getDeclaredMethod("getAttribute", new Class<?>[]{String.class});
      method.setAccessible(true);
      final String result = (String)method.invoke(request, new Object[]{attributeName});
      if (result != null && result.startsWith("//")) {
        return result.substring(1);
      }
      return result;
    }
    catch(final Throwable e) {
      return "<unknown>";
    }
  }
}
