package com.yourkit.probes.builtin;

import com.yourkit.probes.*;

import java.util.WeakHashMap;

/**
 * Probes to monitor monitor some aspects of thread life cycle:
 * where and when threads are created (instance of {@link java.lang.Thread} or its subclass),
 * started ({@link Thread#start()}), and their names changed ({@link Thread#setName(String)})
 * For detail, see <a href="http://www.yourkit.com/docs/95/help/threads_probe.jsp">Help</a>.
 */
public final class Threads {
  /**
   * Mapping between Thread objects and row index in T_THREAD table
   */
  private static final ObjectRowIndexMap<Object> ourThread2RowIndex = new ObjectRowIndexMap<Object>();
  /**
   * Mapping between Thread objects and previous thread name.
   * It is used to avoid redundant name change events.
   */
  private static final WeakHashMap<Object, String> ourThread2PreviousName = new WeakHashMap<Object, String>();

  private static final class ThreadTable extends Table {
    private final StringColumn myName = new StringColumn("Name");

    public ThreadTable() {
      super("Thread Life Cycle", Table.MASK_FOR_GENERIC_LASTING_EVENTS);
    }
  };
  private static final ThreadTable T_THREAD = new ThreadTable();

  private static final class ThreadNameHistoryTable extends Table {
    private final StringColumn myName = new StringColumn("Name");
    private final ForeignKeyColumn myThread = new ForeignKeyColumn(T_THREAD);

    public ThreadNameHistoryTable() {
      super("Name History", Table.MASK_FOR_POINT_EVENTS);
    }
  };
  private static final ThreadNameHistoryTable T_THREAD_NAME_HISTORY = new ThreadNameHistoryTable();

  @MethodPattern("java.lang.Thread:Thread(*)")
  public static final class Thread_constructor_Probe {
    public static void onReturn(
      final @This Object thread
    ) {
      if (!JVM.isLivePhase()) {
        return;
      }

      String name = null;
      try {
        name = ((Thread)thread).getName();
        if (name != null && name.startsWith("YJPAgent-")) {
          // skip profiler internal threads
          return;
        }
      }
      catch (Throwable ignored) {
      }

      final int rowIndex;
      synchronized (ourThread2RowIndex) {
        // Handle nested Thread constructors - create row only once
        final int existingRowIndex = ourThread2RowIndex.get(thread);
        if (existingRowIndex != Table.NO_ROW) {
          rowIndex = existingRowIndex;
        }
        else {
          rowIndex = T_THREAD.createRow();
          ourThread2RowIndex.put(thread, rowIndex);
        }
      }

      handleThreadName(thread, name, rowIndex);
    }
  }

  @MethodPattern("java.lang.Thread:start()")
  public static final class Thread_start_Probe {
    public static void onEnter(final @This Object thread) {
      if (!JVM.isLivePhase()) {
        return;
      }

      final int rowIndex;
      synchronized (ourThread2RowIndex) {
        rowIndex = ourThread2RowIndex.get(thread);
      }
      if (rowIndex == Table.NO_ROW) {
        return;
      }

      T_THREAD.closeRow(rowIndex);
    }
  }

  @MethodPattern("java.lang.Thread:setName(String)")
  public static final class Thread_setName_Probe {
    public static void onReturn(
      final @This Object thread,
      final @Param(1) String name
    ) {
      if (!JVM.isLivePhase()) {
        return;
      }

      final int rowIndex;
      synchronized (ourThread2RowIndex) {
        rowIndex = ourThread2RowIndex.get(thread);
      }
      if (rowIndex == Table.NO_ROW) {
        return;
      }

      handleThreadName(thread, name, rowIndex);
    }
  }

  private static void handleThreadName(
    final Object thread,
    final String newName,
    final int threadRowIndex
  ) {
    if (JVM.isIBM()) {
      // IBM Java sets same name more than once inside thread constructor chain.
      // To get rid of redundant history entries, do nothing if name has not changed.
      synchronized (ourThread2PreviousName) {
        final String previousName = ourThread2PreviousName.get(thread);
        if (previousName != null && previousName.equals(newName)) {
          return;
        }

        ourThread2PreviousName.put(thread, newName);
      }
    }

    // create name history record
    final int historyRowIndex = T_THREAD_NAME_HISTORY.createRow();
    T_THREAD_NAME_HISTORY.myName.setValue(historyRowIndex, newName);
    T_THREAD_NAME_HISTORY.myThread.setValue(historyRowIndex, threadRowIndex);

    // update current thread name
    T_THREAD.myName.setValue(threadRowIndex, newName);
  }
}
